@@ -39,6 +39,9 @@
 
 #include "hfpmodem.h"
 
+#define POLL_CLCC_INTERVAL 2000
+#define CLIP_TIMEOUT 500
+
 static const char *none_prefix[] = { NULL };
 static const char *clcc_prefix[] = { "+CLCC:", NULL };
 
@@ -49,8 +52,9 @@ struct voicecall_data {
 	unsigned int ag_mpty_features;
 	unsigned char cind_pos[HFP_INDICATOR_LAST];
 	int cind_val[HFP_INDICATOR_LAST];
-	unsigned int id_list;
 	unsigned int local_release;
+	unsigned int clcc_source;
+	unsigned int clip_source;
 };
 
 struct release_id_req {
@@ -67,6 +71,8 @@ struct change_state_req {
 	int affected_types;
 };
 
+static gboolean poll_clcc(gpointer user_data);
+
 static GSList *find_dialing(GSList *calls)
 {
 	GSList *c;
@@ -82,10 +88,11 @@ static GSList *find_dialing(GSList *calls)
 	return c;
 }
 
-static struct ofono_call *create_call(struct voicecall_data *d, int type,
+static struct ofono_call *create_call(struct ofono_voicecall *vc, int type,
 					int direction, int status,
 					const char *num, int num_type, int clip)
 {
+	struct voicecall_data *d = ofono_voicecall_get_data(vc);
 	struct ofono_call *call;
 
 	/* Generate a call structure for the waiting call */
@@ -94,7 +101,7 @@ static struct ofono_call *create_call(struct voicecall_data *d, int type,
 	if (!call)
 		return NULL;
 
-	call->id = at_util_alloc_next_id(&d->id_list);
+	call->id = ofono_voicecall_get_next_callid(vc);
 	call->type = type;
 	call->direction = direction;
 	call->status = status;
@@ -116,10 +123,9 @@ static struct ofono_call *new_call_notify(struct ofono_voicecall *vc, int type,
 					int direction, int status,
 					const char *num, int num_type, int clip)
 {
-	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
 	struct ofono_call *c;
 
-	c = create_call(vd, type, direction, status, num, num_type, clip);
+	c = create_call(vc, type, direction, status, num, num_type, clip);
 
 	ofono_voicecall_notify(vc, c);
 
@@ -140,7 +146,6 @@ static void release_call(struct ofono_voicecall *vc, struct ofono_call *call)
 		reason = OFONO_DISCONNECT_REASON_REMOTE_HANGUP;
 
 	ofono_voicecall_disconnected(vc, call->id, reason, NULL);
-	at_util_release_id(&vd->id_list, call->id);
 	vd->local_release &= ~(1 << call->id);
 
 	g_free(call);
@@ -199,6 +204,8 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	GSList *calls;
 	GSList *n, *o;
 	struct ofono_call *nc, *oc;
+	unsigned int num_active = 0;
+	unsigned int num_held = 0;
 
 	dump_response("clcc_poll_cb", ok, result);
 
@@ -214,6 +221,12 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		nc = n ? n->data : NULL;
 		oc = o ? o->data : NULL;
 
+		if (nc && (nc->status == CALL_STATUS_ACTIVE))
+			num_active++;
+
+		if (nc && (nc->status == CALL_STATUS_HELD))
+			num_held++;
+
 		if (oc && (!nc || (nc->id > oc->id))) {
 			enum ofono_disconnect_reason reason;
 
@@ -226,7 +239,6 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 				ofono_voicecall_disconnected(vc, oc->id,
 								reason, NULL);
 
-			at_util_release_id(&vd->id_list, oc->id);
 			vd->local_release &= ~(1 << oc->id);
 
 			o = o->next;
@@ -246,7 +258,8 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 			 */
 			nc->clip_validity = oc->clip_validity;
 
-			if (memcmp(nc, oc, sizeof(struct ofono_call)) && !nc->type)
+			if (memcmp(nc, oc, sizeof(struct ofono_call)) &&
+					!nc->type)
 				ofono_voicecall_notify(vc, nc);
 
 			n = n->next;
@@ -258,6 +271,27 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	g_slist_free(vd->calls);
 
 	vd->calls = calls;
+
+	/* If either active/held call is more than 1, we are in mpty calls.
+	 * we won't get indicator update if any of them is released by CHLD=1x.
+	 * So we have to poll it.
+	 */
+	if (num_active > 1 || num_held > 1)
+		vd->clcc_source = g_timeout_add(POLL_CLCC_INTERVAL, poll_clcc,
+							vc);
+}
+
+static gboolean poll_clcc(gpointer user_data)
+{
+	struct ofono_voicecall *vc = user_data;
+	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
+
+	g_at_chat_send(vd->chat, "AT+CLCC", clcc_prefix,
+				clcc_poll_cb, vc, NULL);
+
+	vd->clcc_source = 0;
+
+	return FALSE;
 }
 
 static void generic_cb(gboolean ok, GAtResult *result, gpointer user_data)
@@ -314,7 +348,7 @@ static void atd_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		ofono_voicecall_notify(vc, call);
 	}
 
-	call = create_call(vd, 0, 0, CALL_STATUS_DIALING, NULL, type, validity);
+	call = create_call(vc, 0, 0, CALL_STATUS_DIALING, NULL, type, validity);
 
 	if (!call) {
 		ofono_error("Unable to allocate call, "
@@ -396,8 +430,8 @@ static void hfp_answer(struct ofono_voicecall *vc,
 static void hfp_hangup(struct ofono_voicecall *vc,
 			ofono_voicecall_cb_t cb, void *data)
 {
-	/* Hangup all calls */
-	hfp_template("AT+CHUP", vc, generic_cb, 0x3f, cb, data);
+	/* Hangup current active call */
+	hfp_template("AT+CHUP", vc, generic_cb, 0x1, cb, data);
 }
 
 static void hfp_hold_all_active(struct ofono_voicecall *vc,
@@ -420,7 +454,8 @@ static void hfp_release_all_held(struct ofono_voicecall *vc,
 	unsigned int held_status = 0x1 << 1;
 
 	if (vd->ag_mpty_features & AG_CHLD_0) {
-		hfp_template("AT+CHLD=0", vc, generic_cb, held_status, cb, data);
+		hfp_template("AT+CHLD=0", vc, generic_cb, held_status,
+				cb, data);
 		return;
 	}
 
@@ -455,6 +490,144 @@ static void hfp_release_all_active(struct ofono_voicecall *vc,
 	CALLBACK_WITH_FAILURE(cb, data);
 }
 
+static void release_id_cb(gboolean ok, GAtResult *result,
+				gpointer user_data)
+{
+	struct release_id_req *req = user_data;
+	struct voicecall_data *vd = ofono_voicecall_get_data(req->vc);
+	struct ofono_error error;
+
+	dump_response("release_id_cb", ok, result);
+	decode_at_error(&error, g_at_result_final_response(result));
+
+	if (ok)
+		vd->local_release |= (1 << req->id);
+
+	req->cb(&error, req->data);
+}
+
+static void hfp_release_specific(struct ofono_voicecall *vc, int id,
+				ofono_voicecall_cb_t cb, void *data)
+{
+	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
+	struct release_id_req *req = NULL;
+	char buf[32];
+
+	if (!(vd->ag_mpty_features & AG_CHLD_1x))
+		goto error;
+
+	req = g_try_new0(struct release_id_req, 1);
+
+	if (!req)
+		goto error;
+
+	req->vc = vc;
+	req->cb = cb;
+	req->data = data;
+	req->id = id;
+
+	sprintf(buf, "AT+CHLD=1%d", id);
+
+	if (g_at_chat_send(vd->chat, buf, none_prefix,
+				release_id_cb, req, g_free) > 0)
+		return;
+
+error:
+	if (req)
+		g_free(req);
+
+	CALLBACK_WITH_FAILURE(cb, data);
+}
+
+static void hfp_private_chat(struct ofono_voicecall *vc, int id,
+				ofono_voicecall_cb_t cb, void *data)
+{
+	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
+	char buf[32];
+
+	if (vd->ag_mpty_features & AG_CHLD_2x) {
+		sprintf(buf, "AT+CHLD=2%d", id);
+
+		hfp_template(buf, vc, generic_cb, 0, cb, data);
+
+		return;
+	}
+
+	CALLBACK_WITH_FAILURE(cb, data);
+}
+
+static void hfp_create_multiparty(struct ofono_voicecall *vc,
+					ofono_voicecall_cb_t cb, void *data)
+{
+	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
+
+	if (vd->ag_mpty_features & AG_CHLD_3) {
+		hfp_template("AT+CHLD=3", vc, generic_cb, 0, cb, data);
+
+		return;
+	}
+
+	CALLBACK_WITH_FAILURE(cb, data);
+}
+
+static void hfp_transfer(struct ofono_voicecall *vc,
+			ofono_voicecall_cb_t cb, void *data)
+{
+	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
+	/* Transfer can puts held & active calls together and disconnects
+	 * from both.  However, some networks support transfering of
+	 * dialing/ringing calls as well.
+	 */
+	unsigned int transfer = 0x1 | 0x2 | 0x4 | 0x8;
+
+	if (vd->ag_mpty_features & AG_CHLD_4) {
+		hfp_template("AT+CHLD=4", vc, generic_cb, transfer, cb, data);
+
+		return;
+	}
+
+	CALLBACK_WITH_FAILURE(cb, data);
+}
+
+static void hfp_send_dtmf(struct ofono_voicecall *vc, const char *dtmf,
+			ofono_voicecall_cb_t cb, void *data)
+{
+	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
+	struct change_state_req *req = g_try_new0(struct change_state_req, 1);
+	char *buf;
+	int s;
+
+	if (!req)
+		goto error;
+
+	req->vc = vc;
+	req->cb = cb;
+	req->data = data;
+	req->affected_types = 0;
+
+	/* strlen("AT+VTS=") = 7 */
+	buf = g_try_new(char, strlen(dtmf) + 7);
+
+	if (!buf)
+		goto error;
+
+	sprintf(buf, "AT+VTS=%s", dtmf);
+
+	s = g_at_chat_send(vd->chat, buf, none_prefix,
+				generic_cb, req, g_free);
+
+	g_free(buf);
+
+	if (s > 0)
+		return;
+
+error:
+	if (req)
+		g_free(req);
+
+	CALLBACK_WITH_FAILURE(cb, data);
+}
+
 static void no_carrier_notify(GAtResult *result, gpointer user_data)
 {
 	DBG("");
@@ -495,16 +668,40 @@ static void ccwa_notify(GAtResult *result, gpointer user_data)
 
 	ofono_debug("ccwa_notify: %s %d %d", num, num_type, validity);
 
-	call = create_call(vd, 0, 1, 5, num, num_type, validity);
+	call = create_call(vc, 0, 1, 5, num, num_type, validity);
 
 	if (!call) {
-		ofono_error("malloc call structfailed. Call management is fubar");
+		ofono_error("malloc call struct failed.  "
+				"Call management is fubar");
 		return;
 	}
 
 	ofono_voicecall_notify(vc, call);
 }
 
+static gboolean clip_timeout(gpointer user_data)
+{
+	struct ofono_voicecall *vc = user_data;
+	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
+	GSList *l;
+	struct ofono_call *call;
+
+	l = g_slist_find_custom(vd->calls,
+				GINT_TO_POINTER(CALL_STATUS_INCOMING),
+				at_util_call_compare_by_status);
+
+	if (l == NULL)
+		return FALSE;
+
+	call = l->data;
+
+	ofono_voicecall_notify(vc, call);
+
+	vd->clip_source = 0;
+
+	return FALSE;
+}
+
 static void ring_notify(GAtResult *result, gpointer user_data)
 {
 	struct ofono_voicecall *vc = user_data;
@@ -544,10 +741,15 @@ static void ring_notify(GAtResult *result, gpointer user_data)
 	}
 
 	/* Generate an incoming call of voice type */
-	call = create_call(vd, 0, 1, CALL_STATUS_INCOMING, NULL, 128, 2);
+	call = create_call(vc, 0, 1, CALL_STATUS_INCOMING, NULL, 128, 2);
 
 	if (!call)
 		ofono_error("Couldn't create call, call management is fubar!");
+
+	/* We don't know the number must wait for CLIP to arrive before
+	 * announcing the call. If timeout, we notify the call as it is.
+	 */
+	vd->clip_source = g_timeout_add(CLIP_TIMEOUT, clip_timeout, vc);
 }
 
 static void clip_notify(GAtResult *result, gpointer user_data)
@@ -604,6 +806,11 @@ static void clip_notify(GAtResult *result, gpointer user_data)
 	call->clip_validity = validity;
 
 	ofono_voicecall_notify(vc, call);
+
+	if (vd->clip_source) {
+		g_source_remove(vd->clip_source);
+		vd->clip_source = 0;
+	}
 }
 
 static void ciev_call_notify(struct ofono_voicecall *vc,
@@ -687,7 +894,7 @@ static void sync_dialing_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	/* Look for dialing or alerting calls on the new list */
 	n = find_dialing(calls);
 
- 	/* Let us find if we have done the dial from HF by looking for
+	/* Let us find if we have done the dial from HF by looking for
 	 * existing dialing or alerting calls
 	 */
 	o = find_dialing(vd->calls);
@@ -803,11 +1010,11 @@ static void ciev_callsetup_notify(struct ofono_voicecall *vc,
 
 		if (o) {
 			struct ofono_call *call = o->data;
-			
+
 			call->status = CALL_STATUS_ALERTING;
 			ofono_voicecall_notify(vc, call);
 		}
-	
+
 		break;
 	}
 
@@ -855,7 +1062,6 @@ static void ciev_callheld_notify(struct ofono_voicecall *vc,
 
 				call->status = CALL_STATUS_HELD;
 				ofono_voicecall_notify(vc, call);
-
 			}
 		} else if (callheld == 1)
 			release_with_status(vc, CALL_STATUS_ACTIVE);
@@ -891,6 +1097,21 @@ static void ciev_notify(GAtResult *result, gpointer user_data)
 		ciev_callheld_notify(vc, value);
 }
 
+static void hfp_clcc_cb(gboolean ok, GAtResult *result, gpointer user_data)
+{
+	struct ofono_voicecall *vc = user_data;
+	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
+	GSList *l;
+
+	if (!ok)
+		return;
+
+	vd->calls = at_util_parse_clcc(result);
+
+	for (l = vd->calls; l; l = l->next)
+		ofono_voicecall_notify(vc, l->data);
+}
+
 static void hfp_voicecall_initialized(gboolean ok, GAtResult *result,
 					gpointer user_data)
 {
@@ -908,6 +1129,9 @@ static void hfp_voicecall_initialized(gboolean ok, GAtResult *result,
 				no_carrier_notify, FALSE, vc, NULL);
 
 	ofono_voicecall_register(vc);
+
+	/* Populate the call list */
+	g_at_chat_send(vd->chat, "AT+CLCC", clcc_prefix, hfp_clcc_cb, vc, NULL);
 }
 
 static int hfp_voicecall_probe(struct ofono_voicecall *vc, unsigned int vendor,
@@ -937,6 +1161,12 @@ static void hfp_voicecall_remove(struct ofono_voicecall *vc)
 {
 	struct voicecall_data *vd = ofono_voicecall_get_data(vc);
 
+	if (vd->clcc_source)
+		g_source_remove(vd->clcc_source);
+
+	if (vd->clip_source)
+		g_source_remove(vd->clip_source);
+
 	g_slist_foreach(vd->calls, (GFunc) g_free, NULL);
 	g_slist_free(vd->calls);
 
@@ -952,18 +1182,17 @@ static struct ofono_voicecall_driver driver = {
 	.dial			= hfp_dial,
 	.answer			= hfp_answer,
 	.hangup			= hfp_hangup,
-	.list_calls		= NULL,
 	.hold_all_active	= hfp_hold_all_active,
 	.release_all_held	= hfp_release_all_held,
 	.set_udub		= hfp_set_udub,
 	.release_all_active	= hfp_release_all_active,
-	.release_specific	= NULL,
-	.private_chat		= NULL,
-	.create_multiparty	= NULL,
-	.transfer		= NULL,
+	.release_specific	= hfp_release_specific,
+	.private_chat		= hfp_private_chat,
+	.create_multiparty	= hfp_create_multiparty,
+	.transfer		= hfp_transfer,
 	.deflect		= NULL,
 	.swap_without_accept	= NULL,
-	.send_tones		= NULL
+	.send_tones		= hfp_send_dtmf
 };
 
 void hfp_voicecall_init()
