@@ -70,6 +70,7 @@ static void huawei_remove(struct ofono_modem *modem)
 
 	ofono_modem_set_data(modem, NULL);
 
+	g_at_chat_unref(data->chat);
 	g_free(data);
 }
 
@@ -113,6 +114,8 @@ static int huawei_enable(struct ofono_modem *modem)
 	if (!data->chat)
 		return -EIO;
 
+	g_at_chat_add_terminator(data->chat, "COMMAND NOT SUPPORT", -1, FALSE);
+
 	if (getenv("OFONO_AT_DEBUG"))
 		g_at_chat_set_debug(data->chat, huawei_debug, NULL);
 
@@ -127,9 +130,14 @@ static int huawei_enable(struct ofono_modem *modem)
 static void cfun_disable(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
+	struct huawei_data *data = ofono_modem_get_data(modem);
 
 	DBG("");
 
+	g_at_chat_shutdown(data->chat);
+	g_at_chat_unref(data->chat);
+	data->chat = NULL;
+
 	if (ok)
 		ofono_modem_set_powered(modem, FALSE);
 }
@@ -143,15 +151,12 @@ static int huawei_disable(struct ofono_modem *modem)
 	if (!data->chat)
 		return 0;
 
+	g_at_chat_cancel_all(data->chat);
+	g_at_chat_unregister_all(data->chat);
 	g_at_chat_send(data->chat, "AT+CFUN=0", NULL,
 					cfun_disable, modem, NULL);
 
-	g_at_chat_shutdown(data->chat);
-
-	g_at_chat_unref(data->chat);
-	data->chat = NULL;
-
-	return 0;
+	return -EINPROGRESS;
 }
 
 static void huawei_pre_sim(struct ofono_modem *modem)
@@ -172,7 +177,7 @@ static void huawei_post_sim(struct ofono_modem *modem)
 	DBG("%p", modem);
 
 	ofono_netreg_create(modem, 0, "atmodem", data->chat);
-	ofono_sms_create(modem, OFONO_VENDOR_HTC_G1, "atmodem", data->chat);
+	ofono_sms_create(modem, OFONO_VENDOR_QUALCOMM_MSM, "atmodem", data->chat);
 }
 
 static struct ofono_modem_driver huawei_driver = {
